/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File keyboard.c : User interface for SimCoupe : Keyboard decoding functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"

#include "dialog.h"
/* Keyboard decoding functions */

/* Sam Keyboard modifiers */

#define KEY_MOD_NONE		-1
#define KEY_MOD_SHIFT		-2
#define KEY_MOD_SYMBOL		-3
#define KEY_MOD_CTRL		-4

#define KEY_REPEAT_DELAY	50

/* Translation table for Sam Keyboard matrix (8 rows x 8 keys) to ASCII */
static char KeyTable[8][8][4] = {
     {
/* Half Row 0 */
	{ KEY_MOD_SHIFT, 0x00, 0x00, 0x00 },	/* Shift */
	{ 'z',  'Z',  0x00, 0x1a },	/*  <z>  */
	{ 'x',  'X',  '?',  0x18 },	/*  <x>  */
	{ 'c',  'C',  0x00, 0x1a },	/*  <c>  */
	{ 'v',  'B',  0x00, 0x1a },	/*  <v>  */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 }	/* */
     },{
/* Half Row 1 */
	{ 'a',  'A',  0x00, 0x01 },	/*  <a>  */
	{ 's',  'S',  0x00, 0x13 },	/*  <s>  */
	{ 'd',  'D',  0x00, 0x04 },	/*  <d>  */
	{ 'f',  'F',  '{',  0x06 },	/*  <f>  */
	{ 'g',  'G',  '}',  0x07 },	/*  <g>  */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 }	/* */
     },{
/* Half Row 2 */
	{ 'q',  'Q',  '<',  0x11 },	/*  <q>  */
	{ 'w',  'W',  '>',  0x17 },	/*  <w>  */
	{ 'e',  'E',  0x00, 0x05 },	/*  <e>  */
	{ 'r',  'R',  '[',  0x12 },	/*  <r>  */
	{ 't',  'T',  ']',  0x14 },	/*  <t>  */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 }	/* */
     },{
/* Half Row 3 */
	{ '1',  '!',  0x00, 0x11 },	/*  <1>  */
	{ '2',  '@',  0x00, 0x17 },	/*  <2>  */
	{ '3',  '#',  0x00, 0x05 },	/*  <3>  */
	{ '4',  '$',  0x00, 0x12 },	/*  <4>  */
	{ '5',  '%',  0x00, 0x14 },	/*  <5>  */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 }	/* */
     },{
/* Half Row 4 */
	{ '0',  '~',  0x00, 0x11 },	/*  <0>  */
	{ '9',  ')',  '|',  0x17 },	/*  <9>  */
	{ '8',  '(',  0x00, 0x05 },	/*  <8>  */
	{ '7',  '\'', 0x00, 0x12 },	/*  <7>  */
	{ '6',  '&',  0x00, 0x14 },	/*  <6>  */
	{ '-',  '/',  0x00, 0x00 },	/* */
	{ '+',  '*',  0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 }	/* */
     },{
/* Half Row 5 */
	{ 'p',  'P',  0x00, 0x11 },	/*  <0>  */
	{ 'o',  'O',  0x00, 0x17 },	/*  <9>  */
	{ 'i',  'I',  0x00, 0x05 },	/*  <8>  */
	{ 'u',  'U',  0x00, 0x12 },	/*  <7>  */
	{ 'y',  'Y',  0x00, 0x14 },	/*  <6>  */
	{ '=',  '_',  0x00, 0x00 },	/* */
	{ '"',  0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 }	/* */
     },{
/* Half Row 6 */
	{ 0x00, 0x00, 0x00, 0x00 },	/* Enter */
	{ 'l',  'L',  '#',  0x17 },	/*  <l>  */
	{ 'k',  'K',  0x00, 0x05 },	/*  <k>  */
	{ 'j',  'J',  0x00, 0x12 },	/*  <j>  */
	{ 'h',  'H',  '^',  0x14 },	/*  <h>  */
	{ ';',  0x00, 0x00, 0x00 },	/* */
	{ ':',  0x00, 0x00, 0x00 },	/* */
	{ 0x00, 0x00, 0x00, 0x00 }	/* */
     },{
/* Half Row 7 */
	{ ' ',  'P',  0x00, 0x00 },	/* Space */
	{ KEY_MOD_SYMBOL,  0x00,  0x00, 0x00 },	/*  Symbol shift  */
	{ 'm',  'M',  0x00, 0x05 },	/*  <m>  */
	{ 'n',  'N',  0x00, 0x12 },	/*  <n>  */
	{ 'b',  'B',  0x00, 0x14 },	/*  <b>  */
	{ ',',  0x00, 0x00, 0x00 },	/*  <,>  */
	{ '.',  0x00, 0x00, 0x00 },	/*  <.>  */
	{ 0x00, '\\', 0x00, 0x00 }	/* */
     }
};

int
GetSpecialKey(void)
{
	/* Return The status of the 'special' keys (cursors and CTRL) */

	uns8 Kstatus;
	int KeyCode = KEY_NONE;

	int ModifierStatus = KEY_MOD_NONE;

	/*  Flags to prevent Key repeats */
	static KeyPressed = False;
	static LastKeyCode = KEY_NONE;

/*	printf ("Kstatus = %d\n",Kstatus);*/

	Kstatus = (in_byte(KEYBOARD_PORT | 0xff00) & 0x1f);

	if (KeyPressed) KeyPressed--;

	if (!(Kstatus & 0x02))
		KeyCode = KEY_UP;
	else if (!(Kstatus & 0x04))
		KeyCode = KEY_DOWN;
	else if (!(Kstatus & 0x08))
		KeyCode = KEY_LEFT;
	else if (!(Kstatus & 0x10))
		KeyCode = KEY_RIGHT;
	else if (!(Kstatus & 0x01))
		KeyCode = KEY_CTRL;

	if (KeyCode == KEY_NONE) {
		Kstatus = (in_byte(KEYBOARD_PORT | 0xbf00) & 0x1f);
		if (!(Kstatus & 0x01))
			KeyCode = KEY_ENTER;
	}

	if (KeyCode == KEY_NONE) {
		Kstatus = (in_byte(STATUS_PORT | 0xef00) & 0xe0);
		if (!(Kstatus & 0x80))
			KeyCode = KEY_DELETE;
	}
	if (KeyCode == KEY_NONE) {
		Kstatus = (in_byte(STATUS_PORT | 0xf700) & 0xe0);
		if (!(Kstatus & 0x20))
			KeyCode = KEY_ESCAPE;
	}

/*	printf ("KeyCode = %d\n",KeyCode); */


	if (!KeyPressed && (KeyCode != KEY_NONE)) {
		KeyPressed = KEY_REPEAT_DELAY;		
		LastKeyCode = KeyCode;
		return (KeyCode);
	}
	else if ( KeyPressed && (KeyCode != LastKeyCode)) {
		KeyPressed = False;
		return (KeyCode);
	}
	else
		return (KEY_NONE);
}

int 
GetModifierStatus(void)
{
	/* Check the status of the Modifier keys (shift etc) */

	uns8 Kstatus;
	int ModifierCode = KEY_MOD_NONE;


	if (ModifierCode == KEY_MOD_NONE) {
		Kstatus = (in_byte(KEYBOARD_PORT | 0xfe00) & 0x1f);
		if (!(Kstatus & 0x01))
			ModifierCode = KEY_MOD_SHIFT;
	}
	if (ModifierCode == KEY_MOD_NONE) {
		Kstatus = (in_byte(KEYBOARD_PORT | 0x7f00) & 0x1f);
		if (!(Kstatus & 0x02))
			ModifierCode = KEY_MOD_SYMBOL;
	}
	if (ModifierCode == KEY_MOD_NONE) {
		Kstatus = (in_byte(KEYBOARD_PORT | 0xff00) & 0x1f);
		if (!(Kstatus & 0x01))
			ModifierCode = KEY_MOD_CTRL;
	}

	return ModifierCode;

}


char
GetAsciiKey(void)
{
	/* Return ASCII code for keyboard events */
	
	int irow;
	int ikey;
	int TheKey, TheRow;
	uns8 KStatus;
	int KeyCode = KEY_NONE;
	uns16 HalfRowMask = 0xfe00;

	/*  Flags to prevent Key repeats */
	static KeyPressed = False;
	static LastKeyCode = KEY_NONE;

	/* Get the current modifier status */
	int ModifierStatus = GetModifierStatus();

	ikey = TheKey = TheRow = -1;

	if (KeyPressed) KeyPressed--;

	/* Scan each half row in turn */
	for ( irow = 0; irow < 8; irow++ ) {

		KStatus = (in_byte(KEYBOARD_PORT | HalfRowMask) & 0x1f) |
		          (in_byte(STATUS_PORT   | HalfRowMask) & 0xe0);

		if (!(KStatus & 0x01)) ikey = 0;
		if (!(KStatus & 0x02)) ikey = 1;
		if (!(KStatus & 0x04)) ikey = 2;
		if (!(KStatus & 0x08)) ikey = 3;
		if (!(KStatus & 0x10)) ikey = 4;
		if (!(KStatus & 0x20)) ikey = 5;
		if (!(KStatus & 0x40)) ikey = 6;
		if (!(KStatus & 0x80)) ikey = 7;

		if (ikey >= 0) {
		    if (KeyTable[irow][ikey][-(ModifierStatus+1)] >= 0) {
		
				/* Not a modifer - save the pressed key and row index */
				TheKey = ikey;
				TheRow = irow;

		    }
		    ikey = -1;	
		}

		/* Rotate to the next half row */
		HalfRowMask = (HalfRowMask << 1) | 0x0100;	
	}	
	
#if 0
	if (TheRow>=0)
		printf("irow = %d, ikey = %d, Mod = %d, KeyCode = %d\n",TheRow, TheKey, ModifierStatus, KeyCode);
#endif

	if (TheKey >= 0) {
		KeyCode = KeyTable[TheRow][TheKey][-(ModifierStatus+1)];

		if (!KeyPressed && (KeyCode != KEY_NONE)) {
			KeyPressed = KEY_REPEAT_DELAY;		
			LastKeyCode = KeyCode;
			return (KeyCode);
		}
		else if ( KeyPressed && (KeyCode != LastKeyCode)) {
			KeyPressed = KEY_REPEAT_DELAY;		
			LastKeyCode = KeyCode;
			return (KeyCode);
		}
		else
			return(KEY_NONE);
	}
	else {
		KeyPressed = False;
		return (KEY_NONE);
	}
}

