/*
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File xevents.c : Event handling code - X11 version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "sam.h"
#include "xsamP.h"
#include "resource.h"
#include "util.h"
#include "version.h"
#include "mouse.h"
#include "xspectrum.h"
#include "fdi.h"
#include "interface.h"

#ifdef XPM_ICON
#include <X11/xpm.h>
#include "simcoupe_icon.xpm"
#else
#include "simcoupe_icon.xbm"
#endif

Display *dpy;
int scrn;
Window root;
Window mainWin;
GC theGc;
XFontStruct *theFont;
Colormap cmap;

static Screen *sptr;
KeySym *key_table;

/* Maps ASCII keys to a keyboard row and bit to set/reset */
/* E.g. ktbl['a'] means set/reset bit 0 (0xfe) of keyboard row 1 */
struct ktbl {
	int port;	/* which keyboard row is the key on? */
	int bit;	/* which bit to mask */
} keytable[128] = {
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	/* SPACE ! "" # $ % & ' */
	{7, 0xfe}, {3, 0xfe}, {9, 0}, {9, 0}, {3,0xf7},
	{3,0xef}, {4,0xf7}, {9, 0},
	/* ( ) * + , - . / */
	{4, 0xfb}, {4, 0xfd}, {4, 0xbf}, {9, 0}, {9, 0}, {9, 0},
	{9, 0}, {9, 0}, 
	/* 0 1 2 3 4 5 6 7 8 9 */
	{4, 0xfe}, {3, 0xfe}, {3, 0xfd}, {3, 0xfb}, {3, 0xf7},
	{3, 0xef}, {4, 0xef}, {4, 0xf7}, {4, 0xfb}, {4, 0xfd},
	/* : ; < = > ? @ */
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {3, 0xfd}, 
	/* A -> Z  (upper & lower case are the same) */
	{1, 0xfe}, {7, 0xef}, {0, 0xf7}, {1, 0xfb}, {2, 0xfb}, {1, 0xf7},
	{1, 0xef}, {6, 0xef}, {5, 0xfb}, {6, 0xf7}, {6, 0xfb}, {6, 0xfd}, 
	{7, 0xfb}, {7, 0xf7}, {5, 0xfd}, {5, 0xfe}, {2, 0xfe}, {2, 0xf7},
	{1, 0xfd}, {2, 0xef}, {5, 0xf7}, {0, 0xef}, {2, 0xfd}, {0, 0xfb},
	{5, 0xef}, {0, 0xfd},
	/* [ \ ] ^ _ ` */
	{9, 0}, {9, 0}, {9, 0}, {4, 0xef}, {9, 0}, {9, 0},
	/* a -> z  (upper & lower case are the same) */
	{1, 0xfe}, {7, 0xef}, {0, 0xf7}, {1, 0xfb}, {2, 0xfb}, {1, 0xf7},
	{1, 0xef}, {6, 0xef}, {5, 0xfb}, {6, 0xf7}, {6, 0xfb}, {6, 0xfd}, 
	{7, 0xfb}, {7, 0xf7}, {5, 0xfd}, {5, 0xfe}, {2, 0xfe}, {2, 0xf7},
	{1, 0xfd}, {2, 0xef}, {5, 0xf7}, {0, 0xef}, {2, 0xfd}, {0, 0xfb},
	{5, 0xef}, {0, 0xfd},
	/* { | } ~ DEL */
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}
};

static struct k_map {
	char *res_name;
	char *def_keysym;
} map_tab[NUM_KEYMAPPINGS] = {
	{ "JoystUp",    "9" }, 
	{ "JoystDown",  "8" },
	{ "JoystLeft",  "6" },
	{ "JoystRight", "7" },
	{ "JoystFire",  "0" },
};

uns8 keyports[10];	/* 8 rows of keys (+ 1 row for unscanned keys) */
uns8 inPorts[256], outPorts[256];

static int xpos, ypos;

/* All to do with screen painting */

int ClutChangeInFrame;			/* signals a clut change has been made in 
					   the present frame */
int BorderChangeInFrame;		/* signals a border change has been made in
					   the present frame */
int VmprChangeInFrame;			/* signals a vmpr change has been made in
					   the present frame */

int SpectrumKeyboard;			/* Switch for spectrum keyboard
					   mode - see spectrum.c */
static int ScaleFactor;			/* Mode for dynamic scale switching */

#define HSize		(HSize1x1*ScaleFactor)	/* Macros used in scale switching */
#define VSize		(VSize1x1*ScaleFactor)

extern int flash_phase;
extern uns8 LineNeedsUpdating[];

extern void (*line_update)(int);
extern void (*line_update_mode_1)(int);
extern void (*line_update_mode_2)(int);
extern void (*line_update_mode_3)(int);
extern void (*line_update_mode_4)(int);

extern void (*screen_refresh)(void);

void call_vmpr(uns8 val);

#define PointerOriginX  (HSize/2)	/* Initial coordinates of the SAM    */
#define PointerOriginY  (VSize/2)	/* mouse pointer (in X Coordinates)  */
#define PointerButtonsOff  0xff		/* Default Mouse button status (Off) */

int SAMPointerMode = 0;			/* Pointer Attatched and x magnification */
static Cursor SAMNullCursor;		/* Special Null Cursor */			

/* from processor.c */
extern int LineNo;
extern int disass;


/* The SAM ports */

uns8 vmpr = 0x81;
uns8 hmpr = 0x01;
uns8 lmpr = 0x1f;
uns8 midi = 0x00;	
uns8 border = 0x00;
uns32 clut[16];
uns8 keyboard = 0x5f;
uns8 status_reg = 0xff;
uns8 line_int = 0xff;

uns8 hpen = 0x00;
uns8 lpen = 0x00;
        	
uns8 clock_reg = 0xff;

uns8 lepr = 0x00;
uns8 hepr = 0x00;

uns8 attr_reg = 0xff;

/*
 *  Private functions start here
 */

/* Perform a clean close-down. */

static void
closedown(void)
{
        close_image();

/* Free up the Colours */

	XFreeColors(dpy, cmap, pixels, NCOLOURS, 0L);
	XAutoRepeatOn(dpy);
	XCloseDisplay(dpy);
}


static int ShiftIsDown = False;

/* Process a KeyPress event */
static void
process_keypress(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;
	int i;	

	XLookupString(kev, buf, 2, &ks, NULL);

	/* First check for the 'special keys' */
	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] |= 0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] |= 0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] |= 0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] |= 0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] |= 0x01;

	switch (ks) {
		/* Emulator control keys */
		case XK_F1: 
			if (!UiActive) {
				InterruptType = Z80_reset;
				for (i = 0; i < 9; i++)
					keyports[i] = 0xff;
			}
			break;
		case XK_F2:	/* Switch Keyboard mode */
			SpectrumKeyboard = !SpectrumKeyboard;
			break;
		case XK_F3:	/* Reset Mouse / pointer */ 
			if (!SAMPointerMode) {
				XDefineCursor(dpy, mainWin, SAMNullCursor);
				XWarpPointer(dpy, mainWin, mainWin, 0, 0, 0, 0,
					     PointerOriginX, PointerOriginY);

				sam_mouse_button_write(PointerButtonsOff);
				sam_mouse_xy_write(0, 0);	

				if (UiActive) {
				    ui_set_button_state(PointerButtonsOff);
				    ui_set_pointer_position(0, 0);
					ui_pointer_draw();
				}
			}
			SAMPointerMode = (SAMPointerMode + 1) % 3;
			if (!SAMPointerMode) {
				ui_pointer_erase();
				XUndefineCursor(dpy, mainWin);
			}
			break;

		case XK_F4:
			break;
		case XK_F5:	/* Rescale windows */
			ScaleFactor = (ScaleFactor == 1) ? 2 : 1;
			ChangeScaleMode(ScaleFactor);

			/* Call call_vmpr to reset the screen write pointers 
			   and fetch the screen from memory */
			call_vmpr(vmpr);
			break;

		case XK_F6: 
		case XK_F7: 
		case XK_F9:
		case XK_F11:
		case XK_F12:	/* Start a dialog session */
			if (!UiActive)
				ui_main();
			break;

		case XK_F8:	/* Generate an NMI */
			if (!UiActive) {
				InterruptType = Z80_nmi;
			}
			break;
			

		case XK_F10:	/* Quit SimCoupe */
			Quit(0);
			break;

		/* Ordinary SAM keys */
		case XK_Tab:
			keyports[3] &= 0xbf; break;
		case XK_Return:
			keyports[6] &= 0xfe; break;
		case XK_Control_L: case XK_Control_R:
		/* <Shift> <Control> is mapped to Caps Lock */
			if (keyports[0] & 0x01)
			     keyports[8] &= 0xfe;
			else
			     keyports[3] &= 0x7f;
			break;
		case XK_Meta_L:
		case XK_Alt_L:
			keyports[7] &= 0xfd; break;
#ifdef DEC
		/* Use DEC "Compose Character" key for <edit> if available
		   else use RHS Alt/Meta key*/
		case XK_Multi_key: 
#endif
		case XK_Alt_R: case XK_Meta_R: case XK_Mode_switch:
			keyports[6] &= 0x7f; break;
		case XK_BackSpace: case XK_Delete:
			keyports[4] &= 0x7f; break;
		case XK_Escape: case XK_grave:
			keyports[3] &= 0xdf; break;
		case XK_Up:
			keyports[8] &= 0xfd; break;
		case XK_Down:
			keyports[8] &= 0xfb; break;
		case XK_Left:
			keyports[8] &= 0xf7; break;
		case XK_Right:
			keyports[8] &= 0xef; break;
		case XK_minus:
			keyports[4] &= 0xdf; break;
		case XK_underscore:
			keyports[5] &= 0xdf; break;
		case XK_equal:
			keyports[5] &= 0xdf; break;
		case XK_plus:
			keyports[0] |= 0x01; 
			keyports[4] &= 0xbf; break;
		case XK_semicolon:
			keyports[6] &= 0xdf; break;
		case XK_colon:
			keyports[0] |= 0x01; 
			keyports[6] &= 0xbf; break;
		case XK_apostrophe:
			keyports[0] &= 0xfe; 
			keyports[4] &= 0xf7; break;
		case XK_comma:
			keyports[7] &= 0xdf; break;
		case XK_period:
			keyports[7] &= 0xbf; break;
		case XK_less:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xfe; break;
		case XK_greater:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xfd; break;
		case XK_slash:
			keyports[0] &= 0xfe; keyports[4] &= 0xdf; break;
		case XK_question:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[0] &= 0xfb; break;
		case XK_backslash:
			keyports[0] &= 0xfe; keyports[7] &= 0x7f;
			break;
		case XK_bar:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[4] &= 0xfd;
			break;
		case XK_ampersand:
			keyports[4] &=0xef; break;
		case XK_bracketleft:
			keyports[7] &= 0xfd; keyports[2] &= 0xf7;
			break;
		case XK_bracketright:
			keyports[7] &= 0xfd; keyports[2] &= 0xef;
			break;
		case XK_braceleft:
			keyports[0] |= 0x01;
			keyports[7] &= 0xfd; keyports[1] &= 0xf7;
			break;                        
		case XK_braceright:
			keyports[0] |= 0x01;
			keyports[7] &= 0xfd; keyports[1] &= 0xef;
			break;                        
		case XK_quotedbl:
			keyports[0] |= 0x01;
			keyports[5] &= 0xbf;
			break;
		case XK_asciitilde:
			keyports[4] &= 0xfe;
			break;
		case XK_asciicircum:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[6] &= 0xef;
			break;
		/* hash and sterling characters must be treated 
		   separately as their position depends on the 
		   type of keyboard - UK or US */
		case XK_sterling:
			keyports[0] |= 0x01;
			keyports[7] &= 0xfd; keyports[6] &= 0xfd;
			break;
		case XK_numbersign:
			keyports[0] &= 0xfe; keyports[3] &=0xfb;
			break;

		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case XK_KP_0:
			keyports[5] &= 0x7f;
			break;
		case XK_KP_1:
			keyports[0] &= 0xdf;
			break;
		case XK_KP_2:
			keyports[0] &= 0xbf;
			break;
		case XK_KP_3:
			keyports[0] &= 0x7f;
			break;
		case XK_KP_4:
			keyports[1] &= 0xdf;
			break;
		case XK_KP_5:
			keyports[1] &= 0xbf;
			break;
		case XK_KP_6:
			keyports[1] &= 0x7f;
			break;
		case XK_KP_7:
			keyports[2] &= 0xdf;
			break;
		case XK_KP_8:
			keyports[2] &= 0xbf;
			break;
		case XK_KP_9:
			keyports[2] &= 0x7f;
			break;
                
                /* Finally the <inv> and (c) keys are mapped onto the
                   keypad <enter> and <.> */
                case XK_KP_Enter:
                        keyports[7] &= 0x7f;
                	break;
                case XK_KP_Decimal:
			keyports[0] &= 0xfe; keyports[5] &= 0xbf;
                	break;

		case XK_Shift_L: case XK_Shift_R:
			keyports[0] &= 0xfe; 
			ShiftIsDown = True;
			break;


		default:
			if (ks < 128) {
				keyports[keytable[ks].port] &= keytable[ks].bit;
			}
			break;
	}
}

/* Process a KeyRelease event */
static void
process_keyrelease(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;

	XLookupString(kev, buf, 2, &ks, NULL);

	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] &= ~0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] &= ~0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] &= ~0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] &= ~0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] &= ~0x01;

	switch (ks) {

		/* Ordinary SAM keys */
		case XK_Tab:
			keyports[3] |= ~0xbf; break;

		case XK_Return:
			keyports[6] |= ~0xfe; break;
		case XK_Control_L: case XK_Control_R:
		/* <Shift> <Control> is mapped to Caps Lock */
			keyports[8] |= ~0xfe;
			keyports[3] |= ~0x7f;
			break;

		case XK_Meta_L:
		case XK_Alt_L: 
			keyports[7] |= ~0xfd; break;
#ifdef DEC
		/* Use DEC "Compose Character" key for <edit> if available
		   else use RHS Alt/Meta key*/
		case XK_Multi_key: 
#endif
		case XK_Alt_R: case XK_Meta_R: case XK_Mode_switch:
			keyports[6] |= ~0x7f; break;
		case XK_BackSpace: case XK_Delete:
			keyports[4] |= ~0x7f; break;
		case XK_Escape: case XK_grave:
			keyports[3] |= ~0xdf; break;
		case XK_Up:
			keyports[8] |= ~0xfd; break;
		case XK_Down:
			keyports[8] |= ~0xfb; break;
		case XK_Left:
			keyports[8] |= ~0xf7; break;
		case XK_Right:
			keyports[8] |= ~0xef; break;
		case XK_minus:
			keyports[4] |= ~0xdf; break;
		case XK_underscore:
			keyports[5] |= ~0xdf; break;
		case XK_equal:
			keyports[5] |= ~0xdf; break;
		case XK_plus:
			keyports[0] &= ~0x01; 
			keyports[4] |= ~0xbf; break;
		case XK_semicolon:
			keyports[6] |= ~0xdf; break;
		case XK_colon:
			keyports[0] &= ~0x01; 
			keyports[6] |= ~0xbf; break;
		case XK_apostrophe:
			keyports[0] |= ~0xfe; 
			keyports[4] |= ~0xf7; break;
		case XK_comma:
			keyports[7] |= ~0xdf; break;
		case XK_period:
			keyports[7] |= ~0xbf; break;
		case XK_less:
			keyports[0] &= ~0x01; 
			keyports[7] |= ~0xfd;
			keyports[2] |= ~0xfe; break;
		case XK_greater:
			keyports[0] &= ~0x01; 
			keyports[7] |= ~0xfd;
			keyports[2] |= ~0xfd; break;
		case XK_slash:
			keyports[0] |= ~0xfe; 
			keyports[4] |= ~0xdf; break;
		case XK_question:
			keyports[0] &= ~0x01; 
			keyports[7] |= ~0xfd;
			keyports[0] |= ~0xfb; break;
		case XK_backslash:
			keyports[0] |= ~0xfe; keyports[7] |= ~0x7f;
			break;
		case XK_bar:
			keyports[0] &= ~0x01; 
			keyports[7] |= ~0xfd;
			keyports[4] |= ~0xfd;
			break;
		case XK_ampersand:
			keyports[4] |= ~0xef; break;
		case XK_bracketleft:
			keyports[7] |= ~0xfd; keyports[2] |= ~0xf7;
			break;
		case XK_bracketright:
			keyports[7] |= ~0xfd; keyports[2] |= ~0xef;
			break;
		case XK_braceleft:
			keyports[0] &= ~0x01; 
			keyports[7] |= ~0xfd; keyports[1] |= ~0xf7;
			break;                        
		case XK_braceright:
			keyports[0] &= ~0x01; 
			keyports[7] |= ~0xfd; keyports[1] |= ~0xef;
			break;                        
		case XK_quotedbl:
			keyports[0] &= ~0x01; 
			keyports[5] |= ~0xbf;
			break;
		case XK_asciitilde:
			keyports[4] |= ~0xfe;
			break;
		case XK_asciicircum:
			keyports[0] &= ~0x01; 
			keyports[7] |= ~0xfd;
			keyports[6] |= ~0xef;
			break;
		/* hash and sterling characters must be treated 
		   separately as their position depends on the 
		   type of keyboard - UK or US */
		case XK_sterling:
			keyports[0] &= ~0x01;
			keyports[7] |= ~0xfd; keyports[6] |= ~0xfd;
			break;
		case XK_numbersign:
			if (!ShiftIsDown) keyports[0] |= 0x01;
			keyports[3] |= ~0xfb;
			break;


		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case XK_KP_0:
			keyports[5] |= ~0x7f;
			break;
		case XK_KP_1:
			keyports[0] |= ~0xdf;
			break;
		case XK_KP_2:
			keyports[0] |= ~0xbf;
			break;
		case XK_KP_3:
			keyports[0] |= ~0x7f;
			break;
		case XK_KP_4:
			keyports[1] |= ~0xdf;
			break;
		case XK_KP_5:
			keyports[1] |= ~0xbf;
			break;
		case XK_KP_6:
			keyports[1] |= ~0x7f;
			break;
		case XK_KP_7:
			keyports[2] |= ~0xdf;
			break;
		case XK_KP_8:
			keyports[2] |= ~0xbf;
			break;
		case XK_KP_9:
			keyports[2] |= ~0x7f;
			break;
                
		/* Finally the <inv> and (c) keys are mapped onto the
		   keypad <enter> and <.> */
		case XK_KP_Enter:
			keyports[7] |= ~0x7f;
			break;
		case XK_KP_Decimal:
			keyports[0] |= ~0xfe; keyports[5] |= ~0xbf;
			break;
		case XK_Shift_L: case XK_Shift_R:
			keyports[0] |= ~0xfe;
			ShiftIsDown = False;
			break;
		default:
			if (ks < 128)
				keyports[keytable[ks].port] |= ~keytable[ks].bit;
			break;
	}
}


/* Process a pointer MotionNotify event */
static void
process_pointermove(XPointerMovedEvent *pev)
{
	int DeltaX, DeltaY;

	DeltaX = ((pev->x - PointerOriginX) * SAMPointerMode) / ScaleFactor;
	DeltaY = (PointerOriginY - pev->y) / ScaleFactor;

	sam_mouse_xy_write(DeltaX, DeltaY);	
	XWarpPointer(dpy, mainWin, mainWin, 0, 0, 0, 0, PointerOriginX, PointerOriginY);

	/* Update the pointer position used for the UI if active */
		
	/* Update the UI pointer position if active */
	if (UiActive)
		ui_set_pointer_position(DeltaX, -DeltaY);
}

/* Process mouse ButtonPressed event */
static void
process_buttonpress(XButtonPressedEvent *bev)
{
	int SAMButtons = PointerButtonsOff;

	switch (bev->button) {

		case Button1:
			SAMButtons &= ~SAM_BUTTON_1;
			break;				
		case Button2:
			SAMButtons &= ~SAM_BUTTON_2;
			break;				
		case Button3:
			SAMButtons &= ~SAM_BUTTON_3;
			break;
		case Button4:
			SAMButtons &= ~SAM_BUTTON_4;
			break;				

	} /* end switch */
	sam_mouse_button_write(SAMButtons);

	/* Update the UI Buttons state if active */
	if (UiActive)
		ui_set_button_state(SAMButtons);

	
}

/* Process mouse ButtonReleased event */
static void
process_buttonrelease(XButtonPressedEvent *bev)
{
	int SAMButtons = PointerButtonsOff;

	switch (bev->button) {

		case Button1:
			SAMButtons |= SAM_BUTTON_1;
			break;				
		case Button2:
			SAMButtons |= SAM_BUTTON_2;
			break;				
		case Button3:
			SAMButtons |= SAM_BUTTON_3;
			break;
		case Button4:
			SAMButtons |= SAM_BUTTON_4;
			break;				
	} /* end switch */
	sam_mouse_button_write(SAMButtons);

	/* Update the UI Buttons state if active */
	if (UiActive)
		ui_set_button_state(SAMButtons);

}


/*
 * Initialise the 128 colours required by the SAM
 * Return -1 if colour allocation failed.
 */
int
get_colours(void)
{
	XColor xc;
	int i;
        int red_index, green_index, blue_index;

        int RGBvals[8] = {0x0000,0x0c30, 0x30c3, 0x3cf3, 
        	          0xc30c,0xcf3c, 0xf3cf, 0xffff}; 

	xc.flags = DoRed | DoGreen | DoBlue;
	for (i = 0; i < NCOLOURS; i++) {

		blue_index  = (i & 0x01) << 1| ((i & 0x10) >> 2) | ((i & 0x08) >> 3);                       
		red_index   = (i & 0x02)     | ((i & 0x20) >> 3) | ((i & 0x08) >> 3);                       
		green_index = (i & 0x04) >> 1| ((i & 0x40) >> 4) | ((i & 0x08) >> 3);                       

	 	xc.red   = RGBvals[red_index];
 		xc.green = RGBvals[green_index];
 		xc.blue  = RGBvals[blue_index];

		if (!XAllocColor(dpy, cmap, &xc)) {
			simcoupe_mesg(SIMCOUPE_WARN, "XAllocColor failed on colour %d", i);
			return(-1);
		}
		pixels[i] = xc.pixel;
	}
	return 0;
}

Cursor CreateNullCursor(Window win)
{
	static char NullCursor_bits[] = { 0x00 };
	Cursor NullCursor;
	Pixmap NullPixmap;
	XColor NullColour[2];
	
	NullPixmap = XCreatePixmapFromBitmapData(dpy,
		win, NullCursor_bits, 1, 1,
		BlackPixel(dpy, scrn), BlackPixel(dpy, scrn),
		1);

	NullColour[0].pixel = BlackPixel(dpy,scrn);
	NullColour[1].pixel = BlackPixel(dpy,scrn);
	XQueryColors(dpy, DefaultColormap(dpy,scrn), NullColour, 2);

	NullCursor = XCreatePixmapCursor(dpy, NullPixmap, NullPixmap, 
		     &NullColour[0], &NullColour[1], 0, 0);

	return(NullCursor);
}

/* Tell the window manager everything it needs to know about the 
 * top-level window.
 */
static void
notify_wm(void)
{
	Pixmap icon_pix;
#ifdef XPM_ICON
	Pixmap shape_pix;
	XpmAttributes icon_xpm_attrs;
#endif
	XWMHints xwmh;
	XSizeHints xsh;
	XClassHint xch;
	XTextProperty appName, iconName;
	char *app_name;
	char *icon_name="SimCoupe";

#ifdef XPM_ICON

	icon_xpm_attrs.visual    = DefaultVisual(dpy, scrn);
	icon_xpm_attrs.colormap  = DefaultColormap(dpy,scrn);
	icon_xpm_attrs.valuemask = XpmCloseness | XpmColormap | XpmVisual;
	XpmCreatePixmapFromData(dpy, root, simcoupe_icon, 
				&icon_pix, &shape_pix, 
				&icon_xpm_attrs);
#else
	icon_pix = XCreatePixmapFromBitmapData(dpy,
		root, simcoupe_icon_bits, simcoupe_icon_width, 
		simcoupe_icon_height,
		BlackPixel(dpy, scrn), WhitePixel(dpy, scrn),
		DefaultDepth(dpy, scrn));

#endif
	app_name = (char *)malloc(20);
	if (!app_name)
		app_name = "simcoupe";
	else
		sprintf(app_name, "simcoupe %d.%d", SIMCOUPE_Major, SIMCOUPE_Minor);

	xsh.flags = PSize | PMinSize | PMaxSize;

	if (strlen(GetCfg(geometry)) > 0)
		xsh.flags |= USPosition;
	else
		xsh.flags |= PPosition;

	xsh.min_width  = HSize1x1; 
	xsh.min_height = VSize1x1;
	xsh.max_width  = HSize2x2;
	xsh.max_height = VSize2x2;

	if (XStringListToTextProperty(&app_name, 1, &appName) == 0) {
		simcoupe_mesg(SIMCOUPE_ERR, "can't create a TextProperty!");
		return;
	}
	if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
		simcoupe_mesg(SIMCOUPE_ERR, "can't create a TextProperty!");
		return;
	}

	xwmh.initial_state = NormalState;
	xwmh.input = True;
	xwmh.icon_pixmap = icon_pix;
	xwmh.flags = StateHint | IconPixmapHint | InputHint;
#ifdef XPM_ICON
	xwmh.icon_mask = shape_pix;
	xwmh.flags |= IconMaskHint;
#endif
	
	xch.res_name = "simcoupe";
	xch.res_class = "Simcoupe";

	XSetWMProperties(dpy, mainWin, &appName, &iconName, 
		&app_name, 1, &xsh, &xwmh, &xch);
}


/* Calculate initial window position depending on the geometry spec.
 * given in the resources
 */
static void
calc_win_position(int *x_ret, int *y_ret)
{
	int x_tmp, y_tmp, w_ret, h_ret;
	int mask;

	if (!GetCfg(geometry)[0]) return;

	mask = XParseGeometry(GetCfg(geometry), &x_tmp, &y_tmp, &w_ret, &h_ret);

	*x_ret = 0; *y_ret = 0;

	if (!mask) {
		simcoupe_mesg(SIMCOUPE_WARN, "invalid geometry specification %s", 
			GetCfg(geometry));
		GetCfg(geometry)[0] = '\0';
		return;
	}

	if (mask & XValue) 
		*x_ret = (mask & XNegative) ?
			(DisplayWidth(dpy, scrn) +  x_tmp -  HSize) :
			x_tmp;

	if (mask & YValue) 
		*y_ret = (mask & YNegative) ?
			(DisplayHeight(dpy, scrn) + y_tmp - VSize) :
			y_tmp;
}

/*
 *  Public functions start here
 */

void
init_display(void)
{
	int i;

	/* Initialise the keysym -> SAM action translations */

	key_table = (KeySym *)malloc(sizeof(KeySym) * NUM_KEYMAPPINGS);
	for (i = 0; i < NUM_KEYMAPPINGS; i++) {
		key_table[i] = XStringToKeysym(map_tab[i].def_keysym);
	}


	/* Connect to the X display */
	if ((dpy = XOpenDisplay(GetCfg(display))) == NULL) {
		simcoupe_mesg(SIMCOUPE_FATAL, "failed to open X display");
	}

	flash_phase = !GetCfg(flashing);

	/* Some useful X variables */
	scrn = DefaultScreen(dpy);
	sptr = DefaultScreenOfDisplay(dpy);
	root = DefaultRootWindow(dpy);

	/* Before we go anywhere we had better check the depth of the
	   visual. */
	if (init_display_depth(DefaultDepth(dpy, scrn))) {
		
		simcoupe_mesg(SIMCOUPE_INFO, "Using %d bit display functions",DefaultDepth(dpy, scrn));
	
	}
	else {
		simcoupe_mesg(SIMCOUPE_FATAL, "SimCoupe only supports visuals with 8,16 or 32 bits per pixel");
	}

	/* Graphics contexts */

	theGc = XCreateGC(dpy, root, 0, NULL);
	XCopyGC(dpy, DefaultGC(dpy, scrn), ~0, theGc);
	XSetGraphicsExposures(dpy, theGc, False);

	/* Colour initialisation */
	if (GetCfg(priv_cmap))
		cmap = XCreateColormap(dpy, root, DefaultVisual(dpy, scrn), AllocNone);
	else
		cmap  = DefaultColormap(dpy, scrn);

	if (get_colours() == -1) {
		simcoupe_mesg(SIMCOUPE_WARN, "colour allocation failed - using private cmap");
		SetCfg(priv_cmap, True);
		cmap = XCreateColormap(dpy, root, DefaultVisual(dpy, scrn), AllocNone);
		if (get_colours() == -1) 
			simcoupe_mesg(SIMCOUPE_FATAL, "private colourmap initialisation failed!");
	}

	/* Initialise variables used to keep track of line interrupt 
	   port changes */

	ClutChangeInFrame   = 0;
	BorderChangeInFrame = 0;
	VmprChangeInFrame   = 0;	

	/* Get a font to use */
	theFont = XLoadQueryFont(dpy, GetCfg(font));
	if (!theFont) {
		simcoupe_mesg(SIMCOUPE_WARN, "can't load font <%s>, trying <%s>...",
			GetCfg(font), DEF_FONT);
		theFont = XLoadQueryFont(dpy, DEF_FONT);
		if (!theFont)
			simcoupe_mesg(SIMCOUPE_FATAL, "can't load font <%s>!", DEF_FONT);
	}
	XSetFont(dpy, theGc, theFont->fid);

	/* Image handling initialisation */
	if (open_image() != 0) {
#ifdef MITSHM
		simcoupe_mesg(SIMCOUPE_WARN,
				"MIT-SHM initialisation failure, trying without MIT-SHM...");
		SetCfg(use_mitshm, False);
#endif
		if (open_image() != 0)
			simcoupe_mesg(SIMCOUPE_FATAL, "image initialisation failure!");
	}


	/* No keys initially being pressed */
	for (i = 0; i < 9; i++)
		keyports[i] = 0xff;
	/* Keyboard starts off in SAM mode by default */
	SpectrumKeyboard = 0;

/* Audio initialisation would go here, but none for now */ 

#ifdef JOY
	/* Linux analogue joystick initialisation */
	if (GetCfg(use_js) && joystick_init() != 0) { 
		simcoupe_mesg(SIMCOUPE_WARN, "joystick initialisation failed");
		SetCfg(use_js, 0);
	}
#endif

	/* Mouse initialisation */
	sam_mouse_init();
	
	calc_win_position(&xpos, &ypos);

	/* Window creation */

	mainWin = XCreateSimpleWindow(dpy, root, xpos, ypos,
			HSize1x1, VSize1x1, 0, pixels[0], pixels[0]);
	
	/* Initialise window and events */

	/* let's be environmentally friendly */
	notify_wm();

	if (GetCfg(priv_cmap))
		XSetWindowColormap(dpy, mainWin, cmap);
	
	XSelectInput(dpy, mainWin, E_MASK | E_MASK_K);

	XMapRaised(dpy, mainWin);

	/* Create a blank X cursor for when the SAM mouse is active */
	SAMNullCursor = CreateNullCursor(mainWin);

	/* Set the dynamic scaling to 1x1 initially */

	ScaleFactor = 1;
	ChangeScaleMode(ScaleFactor);
	line_update = line_update_mode_1;

	XFlush(dpy);

	OnQuit(closedown);
}


/* Check & process any incoming events */
void 
check_events(void)
{
	static XEvent xev;
	static long KeyMask = KeyPressMask;

	XPointerMovedEvent *PointerEvent;

	static FirstTime = True;

	/* If this is the first call to check_events(), after */
	/* the first emulation frame, then use the UI to put  */
	/* up a splach banner.				      */

	if (FirstTime) {
		FirstTime = False;
		ui_banner();
	}

	PointerEvent	= NULL;

#if 1
	while (XEventsQueued(dpy, QueuedAfterReading)) {
		XNextEvent(dpy, &xev);
#else
	while (XCheckWindowEvent (dpy, mainWin, E_MASK | KeyMask, &xev)) {
#endif
		switch(xev.type) {
			case Expose:
				/* force a refresh of the entire screen */
				screen_refresh();
				break;
			case ConfigureNotify:
				XResizeWindow(dpy, mainWin, HSize, VSize);
				break;
			case KeyPress:
				if (!SpectrumKeyboard)
					process_keypress((XKeyEvent *)&xev);
				else	
					process_spectrum_keypress((XKeyEvent *)&xev);
				break;
			case KeyRelease:
				if (!SpectrumKeyboard)
					process_keyrelease((XKeyEvent *)&xev);
				else	
					process_spectrum_keyrelease((XKeyEvent *)&xev);
				break;
			case FocusOut:
				XAutoRepeatOn(dpy);
				break;
			case FocusIn:
				XAutoRepeatOff(dpy);
				break;
			case MotionNotify:
				PointerEvent = (XPointerMovedEvent *)&xev;
				break;
			case ButtonPress:
				process_buttonpress((XButtonPressedEvent *)&xev);
				break;
			case ButtonRelease:
				process_buttonrelease((XButtonReleasedEvent *)&xev);
				break;
			case MapNotify: case UnmapNotify: case ReparentNotify:
				/* just avoid complaints about unhandled events here */
				break;
			default:
				simcoupe_mesg(SIMCOUPE_WARN, "unhandled X event, type %d", xev.type);
				break;
		}
	}

	/* Flip between trapping key press and release events for alternate */
	/* calls to check_events() so we don't lose key events before the   */
	/* emulator can read the ports					    */
	
	KeyMask = (KeyMask == KeyPressMask) ? KeyReleaseMask : KeyPressMask;

	/* We only process the last MotionNotify event - so the SAM mouse
	   code can keep up! */

	if (PointerEvent != NULL && SAMPointerMode) {
		process_pointermove(PointerEvent); 
		PointerEvent = NULL;
	}

	/* If the UI is active call its event handler*/
	
	if (UiActive)
		ui_event();


}