/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File xfb8.c : SAM display driver routines - X11 8 bit Colour version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static uns32 *XMemoryPtr_32;	/* Pointer within the X frame buffer  */

uns32 BorderMemory_32[BORDER_PIXELS+1];

static uns32 *SamBorderPtr_32;	/* Pointer to the SAM border 'memory' */

/* inline functions for mem* type functions to deal */
/* with 32 bit numbers rather than 8 bit ones       */
 
inline void *
memset32(uns32 *s, uns32 c, size_t n)
{
	while (n--)
		*s = c;
	return s;

}

inline void *
memcpy32(uns32 *s, const uns32 *ct, size_t n)
{
	while (n--)
		*s = *ct;
	return s;
}

void border_change_32(void)
{
   long BordPos = LineCycleCounter; /* + 4;*/
      
   BorderChangeInLine = True;

   SamBorderPtr_32 = BorderMemory_32+(BordPos>>3);
   memset32(SamBorderPtr_32, clut[BORD_VAL(border)], BORDER_PIXELS-(BordPos>>3));

}


void
borderline_update_1x1_32(void)
{
/* Update either top or bottom borders if required */

   int i;
   if (ClutChangeInFrame || BorderChangeInFrame) {
#ifdef USE_BORDER_EFFECTS
	if (BorderChangeInLine || True) {
		SamBorderPtr_32 = BorderMemory_32+10;
   		for (i=0; i < BORDER_PIXELS-10; i++) {
			*(XMemoryPtr_32++) = *(SamBorderPtr_32);
			*(XMemoryPtr_32++) = *(SamBorderPtr_32);   		
			*(XMemoryPtr_32++) = *(SamBorderPtr_32);   		
			*(XMemoryPtr_32++) = *(SamBorderPtr_32);
			*(XMemoryPtr_32++) = *(SamBorderPtr_32);   		
			*(XMemoryPtr_32++) = *(SamBorderPtr_32);
			*(XMemoryPtr_32++) = *(SamBorderPtr_32);
			*(XMemoryPtr_32++) = *(SamBorderPtr_32++);
   		}
		XMemoryPtr_32 += HSize1x1;
   	}
   	else {
#else
	{
#endif
	   	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], HSize1x1);
		XMemoryPtr_32 += HSize2x2;
	}
	BorderChangeInLine = False;
   }
   else
	   XMemoryPtr_32 += HSize2x2;
}

void
line_update_x_mode1_1x1_32(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8* SamAttrPtr;
   int i;

   extern unsigned int y_dtbl[];

/* In Mode 1 we have to update SamMemoryPtr each line due to the 
   spectrum screen layout, but we use a lookup table to minimize
   the cost */

   SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
   SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);
   
/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++;
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}

		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr_32 += HSize1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 1 line */
	XMemoryPtr_32 += HSize2x2;	
   }
}

void
line_update_x_mode2_1x1_32(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8* SamAttrPtr;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr once in a given frame, at LineNo == 0 */
   
   if (VmprChangeInFrame || !LineNo)
   	SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	SamAttrPtr = SamMemoryPtr + 0x2000;
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++; 
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}
		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr_32 += HSize1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	XMemoryPtr_32 += HSize2x2;
   }
}


void
line_update_x_mode3_1x1_32(int LineNo)
{
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr_32++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x30) >> 4)];
		*(XMemoryPtr_32++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x03))];
		SamMemoryPtr++;
	}
	
	/* Rhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr_32 += HSize1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 3 line */
	XMemoryPtr_32 += HSize2x2;
   }
}

void
line_update_x_mode4_1x1_32(int LineNo)
{
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
     BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
#ifdef  USE_BORDER_EFFECTS 
	memset32(XMemoryPtr_32, BorderMemory_32[10], BorderWidth_H_1x1);
#else
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1); 
#endif
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr_32++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(XMemoryPtr_32++) = clut[ (*SamMemoryPtr++) &0x0f];
	}
	
	/* Rhs Border */
#ifdef USE_BORDER_EFFECTS
	memset32(XMemoryPtr_32, BorderMemory_32[BORDER_PIXELS-1], BorderWidth_H_1x1);
#else
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_1x1); 
#endif
	XMemoryPtr_32 += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr_32 += HSize1x1;

	LineNeedsUpdating[LineNo] = False;
	BorderChangeInLine = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 4 line */
	XMemoryPtr_32 += HSize2x2;
   }
}


void
borderline_update_2x2_32(void)
{
/* Update either top or bottom borders if required */

   if (ClutChangeInFrame || BorderChangeInFrame)
   	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], 2*HSize2x2);
   
   XMemoryPtr_32 += 2*HSize2x2;
   	
}


void
line_update_x_mode1_2x2_32(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns32 *XMemoryPtrSave;
   uns8* SamAttrPtr;
   int i;

   extern unsigned int y_dtbl[];

/* In Mode 1 we have to update SamMemoryPtr each line due to the 
   spectrum screen layout, but we use a lookup table to minimize
   the cost */

   SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
   SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);
   
/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr_32;

	/* Lhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++;
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}

		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x01) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;


	/* Now generate the 2nd line by copying the first */
	memcpy32(XMemoryPtr_32, XMemoryPtrSave, HSize2x2);
	XMemoryPtr_32 += HSize2x2;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	XMemoryPtr_32 += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}
void
line_update_x_mode2_2x2_32(int LineNo)
{
   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns32 *XMemoryPtrSave;
   uns8* SamAttrPtr;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr once in a given frame, at LineNo == 0 */
   
   if (VmprChangeInFrame || !LineNo)
   	SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   


  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr_32;
   
	/* Lhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	SamAttrPtr = SamMemoryPtr + 0x2000;
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++; 
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}
		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*SamMemoryPtr & 0x01) ? i_pix : p_pix;
		*(XMemoryPtr_32++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;

	/* Now generate the 2nd line by copying the first */
	memcpy32(XMemoryPtr_32, XMemoryPtrSave, HSize2x2);
	XMemoryPtr_32 += HSize2x2;

	LineNeedsUpdating[LineNo] = False;


   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	XMemoryPtr_32 += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}

void
line_update_x_mode3_2x2_32(int LineNo)
{
   uns32 *XMemoryPtrSave;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr_32;

	/* Lhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr_32++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0xc0) >> 6)];
		*(XMemoryPtr_32++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x30) >> 4)];
		*(XMemoryPtr_32++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x0c) >> 2)];
		*(XMemoryPtr_32++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x03))];
		SamMemoryPtr++;
	}
	
	/* Rhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;

	/* Now generate the 2nd line by copying the first */
	memcpy32(XMemoryPtr_32, XMemoryPtrSave, HSize2x2);
	XMemoryPtr_32 += HSize2x2;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 3 line */
	XMemoryPtr_32 += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}

void
line_update_x_mode4_2x2_32(int LineNo)
{
   uns32 *XMemoryPtrSave;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }



/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
     BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   

  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr_32;
   
	/* Lhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr_32++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(XMemoryPtr_32++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(XMemoryPtr_32++) = clut[ (*SamMemoryPtr)   &0x0f];
		*(XMemoryPtr_32++) = clut[ (*SamMemoryPtr++) &0x0f];
	}
	
	/* Rhs Border */
	memset32(XMemoryPtr_32, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr_32 += BorderWidth_H_2x2;


	/* Now generate the 2nd line by copying the first */
	memcpy32(XMemoryPtr_32, XMemoryPtrSave, HSize2x2);
	XMemoryPtr_32 += HSize2x2;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 4 line */
	XMemoryPtr_32 += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}


void
FlyBack_32(void)
{
/* Reset the X frame buffer pointer, and refresh screen */

   XMemoryPtr_32 = x_data_ptr;
   screen_refresh();

}

