/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File xsamgrx.c : SAM display driver routines - X11 version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#if defined(nec_ews) && !defined(__STDC__)
void *shmat(int, void*, int);	/* shut the compiler up */
#endif
#endif

#include "z80.h"
#include "sam.h"
#include "mem.h"
#include "util.h"
#include "xsamP.h"

static void *x_data_ptr;
static XImage *xim;
#ifdef MITSHM
static XShmSegmentInfo xshminfo;
#endif

extern Display *dpy;
extern int scrn;
extern Window root;
extern Window mainWin;
extern GC theGc;

extern int flash_phase;
extern int ClutChangeInFrame;
extern int BorderChangeInFrame;
extern int VmprChangeInFrame;	
extern int BorderChangeInLine;

extern int LineCycleCounter;

/* These macros map a byte in the attribute area to an vga colour index */
#define Background(byte) (((byte) & 0x38) >> 3) | (((byte) & 0x40) >> 3)
#define Foreground(byte) ((byte) & 0x07) | (((byte) & 0x40) >> 3)

uns8 LineNeedsUpdating[256];	/* Determines whether a given line needs to
				   be redrawn */
static uns8 *SamMemoryPtr;	/* Pointer to the SAM screen memory   */

/* General level function pointers */

void (*borderline_update)(void);
void (*line_update)(int);
void (*line_update_mode_1)(int);
void (*line_update_mode_2)(int);
void (*line_update_mode_3)(int);
void (*line_update_mode_4)(int);

void (*screen_refresh)(void);

/* Scale mode level function pointers */

void (*screen_refresh_1x1)(void);
void (*screen_refresh_2x2)(void);

void (*borderline_update_1x1)(void);
void (*borderline_update_2x2)(void);
void (*line_update_x_mode1_1x1)(int);
void (*line_update_x_mode2_1x1)(int);
void (*line_update_x_mode3_1x1)(int);
void (*line_update_x_mode4_1x1)(int);
void (*line_update_x_mode1_2x2)(int);
void (*line_update_x_mode2_2x2)(int);
void (*line_update_x_mode3_2x2)(int);
void (*line_update_x_mode4_2x2)(int);


#ifdef MITSHM
void  screen_refresh_xshm_1x1(void);
void  screen_refresh_xshm_2x2(void);
#endif
void  screen_refresh_xim_1x1(void);
void  screen_refresh_xim_2x2(void);

#define BORDER_PIXELS 48

int BorderTop;
int BorderBottom;

/* Include the Colour depth specific code */

#include "xfb8.c"
#include "xfb16.c"
#include "xfb32.c"

void
ChangeScaleMode(int SF)
{

	XSizeHints xsh;
	int i;
	
	BorderTop    = 35; 
	BorderBottom = 276;
	
	if (SF == 1) {
	
		borderline_update  = borderline_update_1x1;
		line_update_mode_1 = line_update_x_mode1_1x1;
		line_update_mode_2 = line_update_x_mode2_1x1;
		line_update_mode_3 = line_update_x_mode3_1x1;
		line_update_mode_4 = line_update_x_mode4_1x1;

		xsh.min_width  = HSize1x1;
		xsh.min_height = VSize1x1;
		xsh.max_width  = HSize1x1;
		xsh.max_height = VSize1x1;


		screen_refresh       = screen_refresh_1x1;
		XResizeWindow(dpy, mainWin, HSize1x1, VSize1x1);

	}
	else {
	
		borderline_update  = borderline_update_2x2;
		line_update_mode_1 = line_update_x_mode1_2x2;
		line_update_mode_2 = line_update_x_mode2_2x2;
		line_update_mode_3 = line_update_x_mode3_2x2;
		line_update_mode_4 = line_update_x_mode4_2x2;

		xsh.min_width  = HSize2x2; 
		xsh.min_height = VSize2x2;
		xsh.max_width  = HSize2x2;
		xsh.max_height = VSize2x2;

		screen_refresh       = screen_refresh_2x2;
		XResizeWindow(dpy, mainWin, HSize2x2, VSize2x2);

	}

	/* Notify the Window Manager of the change of size */

	xsh.flags = PSize | PMinSize | PMaxSize;
	XSetWMNormalHints(dpy, mainWin, &xsh);	

	/* Make sure we update the whole screen */
	for (i=0; i < 192; i++)
		LineNeedsUpdating[i] = True;
	BorderChangeInFrame = (192+119);
	FlyBack();
}

/*
 * Image routines
 */

#ifdef MITSHM
/* Screen refresh functions - two different functions for each
   scale mode, and one set for either shared or non shared transport */
void 
screen_refresh_xshm_1x1(void)
{

	XShmPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize1x1, VSize1x1, False);
	XFlush(dpy);
}
void 
screen_refresh_xshm_2x2(void)
{

	XShmPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize2x2, VSize2x2, False);
	XFlush(dpy);
}
#endif

void 
screen_refresh_xim_1x1(void)
{

	XPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize1x1, VSize1x1);
	XFlush(dpy);
}
void 
screen_refresh_xim_2x2(void)
{

	XPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize2x2, VSize2x2);
	XFlush(dpy);
}


int
open_image(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		xim = XShmCreateImage(dpy, DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, NULL, &xshminfo, HSize2x2, VSize2x2+8);
		if (!xim) {
			simcoupe_mesg(SIMCOUPE_ERR, "SimCoupe: failed to create a shared XImage!");
			return -1;
		}
		xshminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line * xim->height,
			IPC_CREAT|0777);
		if (xshminfo.shmid == -1) {
			simcoupe_mesg(SIMCOUPE_ERR, "shmget <xshminfo.shmid> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.shmaddr = xim->data = shmat(xshminfo.shmid, 0, 0);
		if (xshminfo.shmaddr == -1) {
			simcoupe_mesg(SIMCOUPE_ERR, "shmat <xshminfo.shmaddr> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.readOnly = False;
		if (!XShmAttach(dpy, &xshminfo)) {
			simcoupe_mesg(SIMCOUPE_ERR, "XShmAttach failed");
			perror("  reason:");
			return -1;
		}
		XSync(dpy, False);
/*		shmctl(xshminfo.shmid, IPC_RMID, (struct shmid_ds *)NULL);*/

		screen_refresh_1x1 = screen_refresh_xshm_1x1;
		screen_refresh_2x2 = screen_refresh_xshm_2x2;
	} else
#endif
	{
		xim = XCreateImage(dpy,  DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, 0, NULL, HSize2x2, VSize2x2+8, 8, 0);
		if (!xim) {
			simcoupe_mesg(SIMCOUPE_ERR, "XCreateImage failed");
			perror("  reason:");
			return -1;
		}
		if ((xim->data = (char *)malloc(xim->bytes_per_line * xim->height))
			== NULL) {
			simcoupe_mesg(SIMCOUPE_ERR, "malloc <xim->data> failed");
			perror("  reason:");
			return -1;
		}
		screen_refresh_1x1 = screen_refresh_xim_1x1;
		screen_refresh_2x2 = screen_refresh_xim_2x2;
	}
	x_data_ptr = xim->data;

	return 0;
}

void
close_image(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		XShmDetach(dpy, &xshminfo);
		XDestroyImage(xim);
		shmdt(xshminfo.shmaddr);
		shmctl(xshminfo.shmid, IPC_RMID, (struct shmid_ds *)NULL);
	}
	else
#endif
		XDestroyImage(xim);
}

int
init_display_depth(int DisplayDepth)
{
	/* Initialise the display related function pointers to the */
	/* correct actual functions for a given depth		   */


	/* NB Only certain depths supported at present */
	
	if (DisplayDepth == 8) {
		line_update_x_mode1_1x1 = line_update_x_mode1_1x1_8; 
		line_update_x_mode2_1x1 = line_update_x_mode2_1x1_8; 
		line_update_x_mode3_1x1 = line_update_x_mode3_1x1_8; 
		line_update_x_mode4_1x1 = line_update_x_mode4_1x1_8; 
		line_update_x_mode1_2x2 = line_update_x_mode1_2x2_8; 
		line_update_x_mode2_2x2 = line_update_x_mode2_2x2_8; 
		line_update_x_mode3_2x2 = line_update_x_mode3_2x2_8; 
		line_update_x_mode4_2x2 = line_update_x_mode4_2x2_8; 

		borderline_update_1x1   = borderline_update_1x1_8; 
		borderline_update_2x2   = borderline_update_2x2_8; 
		border_change = border_change_8;
		FlyBack = FlyBack_8;
	}
	else if (DisplayDepth == 16) {
		line_update_x_mode1_1x1 = line_update_x_mode1_1x1_16; 
		line_update_x_mode2_1x1 = line_update_x_mode2_1x1_16; 
		line_update_x_mode3_1x1 = line_update_x_mode3_1x1_16; 
		line_update_x_mode4_1x1 = line_update_x_mode4_1x1_16; 
		line_update_x_mode1_2x2 = line_update_x_mode1_2x2_16; 
		line_update_x_mode2_2x2 = line_update_x_mode2_2x2_16; 
		line_update_x_mode3_2x2 = line_update_x_mode3_2x2_16; 
		line_update_x_mode4_2x2 = line_update_x_mode4_2x2_16; 

		borderline_update_1x1   = borderline_update_1x1_16; 
		borderline_update_2x2   = borderline_update_2x2_16; 
		border_change = border_change_16;
		FlyBack = FlyBack_16;
	}
	else if (DisplayDepth == 32) {
		line_update_x_mode1_1x1 = line_update_x_mode1_1x1_32; 
		line_update_x_mode2_1x1 = line_update_x_mode2_1x1_32; 
		line_update_x_mode3_1x1 = line_update_x_mode3_1x1_32; 
		line_update_x_mode4_1x1 = line_update_x_mode4_1x1_32; 
		line_update_x_mode1_2x2 = line_update_x_mode1_2x2_32; 
		line_update_x_mode2_2x2 = line_update_x_mode2_2x2_32; 
		line_update_x_mode3_2x2 = line_update_x_mode3_2x2_32; 
		line_update_x_mode4_2x2 = line_update_x_mode4_2x2_32; 

		borderline_update_1x1   = borderline_update_1x1_32; 
		borderline_update_2x2   = borderline_update_2x2_32; 
		border_change = border_change_32;
		FlyBack = FlyBack_32;

	}
	else
		return False;

	return True;

}
